# required packages TreeSim (simulating trees), FossilSim (simulating taxa) and fbdR (estimating speciation and extinction rates)

library(ggplot2)
library(gridExtra)

start.time = Sys.time()

source("../simulation_tests.R")

cat(NULL, file="log.txt", append = F)

set.seed(1234)

if(print.data)
  system("mkdir data")

# simulate trees 1
lambda.branching = 0.2 # branching rate
mu.branching = 0.02 # death rate

print("simulating trees")
trees1<-TreeSim::sim.rateshift.taxa(tips,replicates,
                                    lambda=c(lambda.branching,lambda.branching,0.3),
                                    mu=c(0.3,mu.branching,mu.branching),
                                    frac=c(1,1,1),
                                    times=c(0,15,25))

if(print.data)
  print.trees(trees1, 1)

f = 0 # probability of symmetric speciation
lambda.a = 0 # rate of anagenic speciation

# figure labels
lb1 = "list(lambda==0.2,0.2,0.3, mu==0.3,0.02,0.02)"
lb2 = "list(beta==0, lambda[a]==0)"
lb3 = "list(lambda~\"*\"==0.2,0.2,0.3, mu~\"*\"==0.3,0.02,0.02)"

#ape::ltt.plot(trees1[[1]])

# simulate ranges
ranges = simulate.ranges(trees1)

if(print.data)
  print.ranges(ranges, 1)

# likelihood ratio tests
out = likelihood.ratio.tests(trees1, ranges, "out1.log", shifts = TRUE)

# figures + ouput
p1 = estimates.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE)
d1 = correlations.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE)
c1 = correlations.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE, equation = 2)
o1 = correlations.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE, equation = 3)
out1 = out

# check for convergence issues
convergence.issues(out, file="log.txt")

cat("finished 1...\n", file="log.txt", append=T)

f = 0.5
lambda.a = 0.04

lb1 = "list(lambda==0.2,0.2,0.3, mu==0.3,0.02,0.02)"
lb2 = "list(beta==0.5, lambda[a]==0.04)"
lb3 = "list(lambda~\"*\"==0.34,0.34,0.49, mu~\"*\"==0.44,0.16,0.21)"

ranges = simulate.ranges(trees1)

if(print.data)
  print.ranges(ranges, 2)

out = likelihood.ratio.tests(trees1, ranges, "out2.log", shifts = TRUE)

p2 = estimates.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE)
d2 = correlations.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE)
c2 = correlations.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE, equation = 2)
o2 = correlations.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE, equation = 3)
out2 = out

# check for convergence issues
convergence.issues(out, file="log.txt")

cat("finished 2...\n", file="log.txt", append=T)

f = 0.7
lambda.a = 0.16

lb1 = "list(lambda==0.2,0.2,0.3, mu==0.3,0.02,0.02)"
lb2 = "list(beta==0.7, lambda[a]==0.16)"
lb3 = "list(lambda~\"*\"==0.5,0.5,0.67, mu~\"*\"==0.6,0.32,0.39)"

ranges = simulate.ranges(trees1)

if(print.data)
  print.ranges(ranges, 3)

out = likelihood.ratio.tests(trees1, ranges, "out3.log", shifts = TRUE)

p3 = estimates.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE)
d3 = correlations.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE)
c3 = correlations.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE, equation = 2)
o3 = correlations.figure(out, lb1, lb2, lb3 = lb3, shifts = TRUE, equation = 3)
out3 = out

# check for convergence issues
convergence.issues(out, file="log.txt")

cat("finished 3...\n", file="log.txt", append=T)


##### output

# fig 1

pdf(file="simulations-rate-shift-estimates.pdf",width=12,height=4)
grid.arrange(p1,p2,p3,ncol=3)
dev.off()

# fig 2

pdf(file="simulations-rate-shift-correlations3.pdf",width=12,height=4)
grid.arrange(d1,d2,d3,ncol=3)
dev.off()

# fig 3

pdf(file="simulations-rate-shift-correlations1.pdf",width=12,height=4)
grid.arrange(c1,c2,c3,ncol=3)
dev.off()

# fig 4

pdf(file="simulations-rate-shift-correlations2.pdf",width=12,height=4)
grid.arrange(o1,o2,o3,ncol=3)
dev.off()

# summary file

print.headers()
print.results(out1)
print.results(out2)
print.results(out3)

### 

end.time = Sys.time() - start.time

sink("log.txt", append = T)
cat(end.time, "\n", sep = "", file="log.txt", append=T)
sink()
