# required packages TreeSim (simulating trees), FossilSim v1 (simulating taxa) and fbdR (estimating speciation and extinction rates)
library(ggplot2)
library(gridExtra)

start.time = Sys.time()

source("../simulation_tests.R")

cat(NULL, file="log.txt", append = F)

set.seed(1234)

if(print.data)
  system("mkdir data")

# simulate trees + range trees

trees1 = list()
range.trees1 = list()
lambda.branching = c()
mu.branching = c()
lambda.star = c()
mu.star = c()

for(i in 1:replicates){
  
  birth = runif(1, 0, 1.5)
  death = runif(1, 0, birth*0.98)  
  
  tips = runif(1, 25, 300)
  
  lambda.branching = c(lambda.branching, birth)
  mu.branching = c(mu.branching, death)
  
  t1<-TreeSim::sim.bd.taxa(tips, 1, birth, death)[[1]]
  
  trees1[[i]] = t1
  
  birth = runif(1, 0, 1.5)
  death = runif(1, 0, birth*0.98)
  
  tips = runif(1, 25, 300)
  
  lambda.star = c(lambda.star, birth)
  mu.star = c(mu.star, death)
  
  t2<-TreeSim::sim.bd.taxa(tips, 1, birth, death)[[1]]
  
  range.trees1[[i]] = t2
  
}

lambda.branching = mean(lambda.branching)
mu.branching = mean(mu.branching)
lambda.star = mean(lambda.star)
mu.star = mean(mu.star)

f = 0 # probability of symmetric speciation, if f = 0 all speciation events will be asymmetric, if f = 1 all speciation events will be symmetric
lambda.a = 0 # rate of anagenic speciation

# figure labels
lb1 = "list(lambda==U(0.1, 1.5), mu==U(0.1, lambda), beta==0, lambda[a]==0)"
lb2 = "list(lambda~\"*\"==U(0.1, 1.5),mu~\"*\"==U(0.1, lambda~\"*\"))"

# simulate ranges
ranges = simulate.ranges(range.trees1)

# likelihood ratio tests
out = likelihood.ratio.tests(trees1, ranges, "out1.log")

# figures + ouput
p1 = estimates.figure(out, lb1, lb2, xylim = 1.5)
d1 = correlations.figure(out, lb1, lb2)
c1 = correlations.figure(out, lb1, lb2, equation = 2)
o1 = correlations.figure(out, lb1, lb2, equation = 3)
out1 = out

# check for convergence issues
convergence.issues(out, file="log.txt")

cat("finished 1...\n", file="log.txt", append=T)

##### output

# fig 1

pdf(file="simulations-estimates.pdf",width=4,height=4)
grid.arrange(p1,ncol=1)
dev.off()

# fig 2

pdf(file="simulations-correlations3.pdf",width=4,height=8)
grid.arrange(d1,ncol=1)
dev.off()

# fig 3

pdf(file="simulations-correlations1.pdf",width=4,height=8)
grid.arrange(c1,ncol=1)
dev.off()

# fig 4

pdf(file="simulations-correlations2.pdf",width=4,height=8)
grid.arrange(o1,ncol=1)
dev.off()

# summary file

print.headers()
print.results(out1)

end.time = Sys.time() - start.time

sink("log.txt", append = T)
cat(end.time, "\n", sep = "", file="log.txt", append=T)
sink()

warnings()

# lambda.star = f*2*lambda.branching+(1-f)*lambda.branching+lambda.a
# mu.star = f*lambda.branching+lambda.a+mu.branching

