##### global variables
replicates = 100
recalculate.lk = 10 # 10
attempts.max = 5 # 5
tips = 200
rho = 1
crown = FALSE
debugging = FALSE
print.data = FALSE
# Likelihood ratio test values
ERM_CRM_LR1 = 2.17 # Equal rates vs compatible rates, alpha = 0.05
ERM_CRM_LR2 = 5.41 # Equal rates vs compatible rates, alpha = 0.01
CRM_IRM_LR1 = qchisq(.95, df = 1) # Compatible rates vs incompatible rates, alpha = 0.05
CRM_IRM_LR2 = qchisq(.99, df = 1) # Compatible rates vs incompatible rates, alpha = 0.01
ERM_IRM_LR1 = qchisq(.95, df = 2) # Equal rates vs incompatible rates, alpha = 0.05
ERM_IRM_LR2 = qchisq(.99, df = 2) # Equal rates vs incompatible rates, alpha = 0.01

##### simulate data under the null model

simulate.ranges<-function(trees, parent.labels = FALSE){

  ranges = list()

  for(i in 1:length(trees)){

    tree1<-trees[[i]]

    # assign speciation events as symmetric or asymmetric
    ages <- FossilSim::mixed.ages(tree1, f, root.edge = TRUE)
    # simulate anagenic species
    frs <- FossilSim::anagenic.species(ages, lambda.a, parent.labels = parent.labels)

    ranges[[i]] = frs
  }
  return(ranges)
}

##### poisson sampling [review these functions]

poisson.remove<-function(ranges, r){
  
  for(i in 1:length(ranges)){
    
    frs = ranges[[i]]
  
    length = frs$start - frs$end
  
    prob_keep_range = 1 - exp(-r*length)
  
    frs = frs[which(runif(length(prob_keep_range), 0, 1) < prob_keep_range),]
    
    ranges[[i]] = frs
  }
  return(ranges)
}

poisson.occurrences<-function(ranges, r){
  
  for(i in 1:length(ranges)){
    
    frs = ranges[[i]]
    
    fossils <- data.frame(h = numeric(), sp = numeric())
    
    for(j in 1:length(frs$sp)){
      
      rand = rpois(1, (frs$start[j] - frs$end[j]) * r)
      
      if (rand > 0) {
        h = runif(rand, min = frs$end[j], max = frs$start[j])
        fossils <- rbind(fossils, data.frame(h = h, sp = paste0("taxa_", j)))
      }
      
      # add extant occurrence
      if(frs$end[j] == 0)
        fossils <- rbind(fossils, data.frame(h = 0, sp = paste0("taxa_", j)))
    }
    ranges[[i]] = fossils
  }
  return(ranges)
}

parse.pyrate<-function(occ, prefix = NULL){
  
  file = ""
  
  for(i in 1:length(occ)){
    
    if(is.data.frame(occ)) frs = occ 
    else frs = occ[[i]]
    
    if(!is.null(prefix))
      file = paste0(prefix, "_", i, ".txt")
    
    cat("Species\tStatus\tmin_age\tmax_age\n", file = file, append = FALSE)
    
    for(j in unique(frs$sp)){
      
      h = frs$h[which(frs$sp == j)]
    
      extant = FALSE
    
      if(any(h == 0))
        extant = TRUE
      
      #print(j)
      #print(extant)
      
      if(extant & length(h) == 1)
        next  
      else if (any(h == 0))
        h = h[-which(h == 0)]
      
      for(k in h){
        if(extant)
          cat(j, "\t", "extant", "\t", k, "\t", k, "\n", file = file, append = TRUE)
        else
          cat(j, "\t", "extinct", "\t", k, "\t", k, "\n", file = file, append = TRUE)
      }
    
    }
    
  }
  #eof
}

parse.trees<-function(trees, prefix = NULL, extant.only = FALSE, nexus = TRUE){
  
  file = ""
  
  for(i in 1:length(trees)){
    
    if(is.data.frame(trees)) tree = trees
    else tree = trees[[i]]
    
    if(!is.null(prefix) & nexus)
      file = paste0(prefix, "_", i, ".nex")
    else if(!is.null(prefix))
      file = paste0(prefix, "_", i, ".txt")
    
    if(extant.only)
      tree = geiger::drop.extinct(tree)
    
    if(nexus)
      ape::write.nexus(tree, file = file, translate = TRUE)
    else
      ape::write.tree(tree, file = file)
    
  }
  #eof
}

read.pyrate.ranges<-function(prefix, suffix){
  
  ranges = list()
  
  for(i in 1:replicates){
    
    frs <- read.table(file = paste0(prefix, i, suffix), header=TRUE)
    
    # rename range data headers
    names(frs)[2]<-"sp"
    names(frs)[3]<-"start"
    names(frs)[4]<-"end"
    
    ranges[[i]] = frs
    
  }
  return(ranges)
}

read.trees <-function(prefix){
  
  trees = list()
  
  for(i in 1:replicates){
    
    t <- ape::read.nexus(file = paste0(prefix, i, ".nex"))
    
    trees[[i]] = t
    
  }
  return(trees)
}

##### randomly remove missing data [review these functions]

randomRows<-function(df,n){
  return(df[sample(nrow(df),n),])
}

randomly.remove<-function(ranges, frac){

  missing = frac

  for(i in 1:length(ranges)){

    frs = ranges[[i]]

    # randomly remove fossils
    elimate = ceiling(length(frs$sp) * missing)
    frs = randomRows(frs, length(frs$sp) - elimate)

    ranges[[i]] = frs

  }
  return(ranges)
}

##### non-randomly remove missing data [review these functions]

# non-random sampling
non.random.sampling<-function(ranges, rate){
  
  for(i in 1:length(ranges)){
    
    frs = ranges[[i]]
    
    remove = c()
    
    for(j in 1:length(frs$sp)){
      
      duration = frs$start[j] - frs$end[j]
      
      pr.sampling = 1 - exp(-(rate*duration))
    
      if(runif(1) > pr.sampling) remove = c(remove, j)
        
    }
    frs = frs[-remove,]
    ranges[[i]] = frs 
  }
  return(ranges)
 #eof  
}

# non-randomly remove extant ranges
non.randomly.remove<-function(ranges, frac){
  
  for(i in 1:length(ranges)){
    
    frs = ranges[[i]]
    
    extinct = frs[which(frs$end != 0),]
    extant = frs[which(frs$end == 0),]
    
    # non-randomly remove extant taxa
    elimate = ceiling(length(extant$sp) * frac)
    extant = randomRows(extant, length(extant$sp) - elimate)
    
    frs = rbind(extinct, extant)
    
    ranges[[i]] = frs
    
  }
  return(ranges)
}

# non-randomly remove extinct ranges
non.randomly.remove.backwards<-function(ranges, frac){
  
  for(i in 1:length(ranges)){
    
    frs = ranges[[i]]
    
    extinct = frs[which(frs$end != 0),]
    extant = frs[which(frs$end == 0),]
    
    # non-randomly remove extinct taxa
    elimate = ceiling(length(extinct$sp) * frac)
    extinct = randomRows(extinct, length(extinct$sp) - elimate)
    
    frs = rbind(extinct, extant)
    
    ranges[[i]] = frs
    
  }
  return(ranges)
}

##### cryptic speciation

cryptic.species<-function(ranges, kappa){

  for(i in 1:length(ranges)){

    frs = ranges[[i]]

    # assign cryptic species
    input<-FossilSim::cryptic.speciation(frs, kappa)

    # transform cryptic species assignments into ranges
    sp = c()
    start = c()
    end = c()

    for(j in unique(input$cryptic.id)){
      sp = c(sp, j)
      start = c( start, max(input$start[which(input$cryptic.id == j)]) )
      #end = c(end, max(input$end[which(input$cryptic.id == j)]) )
      end = c(end, min(input$end[which(input$cryptic.id == j)]) )
    }

    frs = data.frame(sp = sp, start = start, end = end)

    ranges[[i]] = frs

  }
  return(ranges)
}

##### likelihood ratio tests

likelihood.ratio.tests<-function(trees, ranges, file = "", shifts = FALSE, file2 = "log.txt"){
  
  ## parameters
  
  ## data to collect
  
  m1.birth.extant = c()
  m1.death.extant = c()
  m1.birth.ranges = c()
  m1.death.ranges = c()
  m1.lk = c()
  m1.convergence = c()
  
  m2.birth.extant = c()
  m2.death.extant = c()
  m2.birth.ranges = c()
  m2.death.ranges = c()
  m2.lk = c()
  m2.convergence = c()
  
  m3.birth.extant = c()
  m3.death.extant = c()
  m3.birth.ranges = c()
  m3.death.ranges = c()
  m3.lk = c()
  m3.convergence = c()
  
  if(debugging)
    cat("Model 1...\n")
  
  ## model 1
  ## lambda = lambda.star and mu = mu.star
  for(i in 1:replicates){
    
    if(debugging)
      cat(i,"\n")
    
    tree1<-trees[[i]]
    frs<-ranges[[i]]
    
    lnL = -10^100
    tmp = NULL
    
    for(j in 1:recalculate.lk){
      
      attempts = 1
      convergence = 1
      
      while(convergence != 0){
        
        b = runif(1)
        d = runif(1, 0, b*.99)
        
        est = fbdR::est.bd.combined(tree1, frs, b = b, d = d, constrained = TRUE, crown = crown, rho = rho, constrained.p = 2)
        
        convergence = est$convergence
        
        if(attempts == attempts.max)
          break
        
        attempts = attempts + 1
      }
      
      if(convergence != 0)
        next
      else{
        if(est$value > lnL){
          lnL = est$value
          tmp = est
        }
      }
    }
    
    # lk estimation has not converged
    if(is.null(tmp)){
      m1.birth.extant = c(m1.birth.extant, NA)
      m1.death.extant = c(m1.death.extant, NA)
      m1.birth.ranges = c(m1.birth.ranges, NA)
      m1.death.ranges = c(m1.death.ranges, NA)
      m1.lk = c(m1.lk, NA)
      m1.convergence = c(m1.convergence, est$convergence)
    } else {
      est = tmp
      m1.birth.extant = c(m1.birth.extant, est$par[1])
      m1.death.extant = c(m1.death.extant, est$par[2])
      m1.birth.ranges = c(m1.birth.ranges, est$par[1])
      m1.death.ranges = c(m1.death.ranges, est$par[2])
      m1.lk = c(m1.lk, est$value)
      m1.convergence = c(m1.convergence, est$convergence)
    }
  }
  
  cat("finished M1...\n", file=file2, append=T)
  
  if(debugging)
    cat("Model 2...\n")
  
  ## model 2
  ## constrained model with lambda, lambda.star and d (= d*)
  for(i in 1:replicates){
    
    if(debugging)
      cat(i,"\n")
    
    tree1<-trees[[i]]
    frs<-ranges[[i]]
    
    lnL = -10^100
    tmp = NULL
    
    for(j in 1:recalculate.lk){
      
      attempts = 1
      convergence = 1
      
      while(convergence != 0){
        
        b = runif(1)
        b.star = runif(1, b, 2)
        nd = runif(1, 0, b*.99)
        
        est = fbdR::est.bd.combined(tree1, frs, b = b, b.star = b.star, nd = nd, constrained = TRUE, crown = crown, rho = rho)
        
        convergence = est$convergence
        
        if(attempts == attempts.max)
          break
        
        attempts = attempts + 1
      }
      
      if(convergence != 0)
        next
      else{
        if(est$value > lnL){
          lnL = est$value
          tmp = est
        }
      }
    }
    
    # lk estimation has not converged
    if(is.null(tmp)){
      m2.birth.extant = c(m2.birth.extant, NA)
      m2.death.extant = c(m2.death.extant, NA)
      m2.birth.ranges = c(m2.birth.ranges, NA)
      m2.death.ranges = c(m2.death.ranges, NA)
      m2.lk = c(m2.lk, NA)
      m2.convergence = c(m2.convergence, est$convergence)
    } else {
      est = tmp
      m2.birth.extant = c(m2.birth.extant, est$par[1])
      m2.birth.ranges = c(m2.birth.ranges, est$par[2])
      m2.death.extant = c(m2.death.extant, est$par[1]-est$par[3])
      m2.death.ranges = c(m2.death.ranges, est$par[2]-est$par[3])
      m2.lk = c(m2.lk, est$value)
      m2.convergence = c(m2.convergence, est$convergence)
    }
  }
  
  cat("finished M2...\n", file=file2, append=T)
  
  if(debugging)
    cat("Model 3...\n")
  
  ## model 3
  ## lambda = lambda.star and mu = mu.star 
  for(i in 1:replicates){
    
    if(debugging)
      cat(i,"\n")
    
    tree1<-trees[[i]]
    frs<-ranges[[i]]
    
    lnL = -10^100
    tmp = NULL
    
    for(j in 1:recalculate.lk){
      
      attempts = 1
      convergence = 1
      
      while(convergence != 0){
        
        b = runif(1)
        d = runif(1, 0, b*.99)
        b.star = runif(1, b, 2)
        d.star = runif(1, 0, b.star*.99)
        
        est = fbdR::est.bd.combined(tree1, frs, b = b, d = d, b.star = b.star, d.star = d.star, constrained = FALSE, crown = crown, rho = rho)
        
        convergence = est$convergence
        
        if(attempts == attempts.max)
          break
        
        attempts = attempts + 1
      }
      
      if(convergence != 0)
        next
      else{
        if(est$value > lnL){
          lnL = est$value
          tmp = est
        }
      }
      
    }
    
    if(is.null(tmp)){
      m3.birth.extant = c(m3.birth.extant, NA)
      m3.death.extant = c(m3.death.extant, NA)
      m3.birth.ranges = c(m3.birth.ranges, NA)
      m3.death.ranges = c(m3.death.ranges, NA)
      m3.lk = c(m3.lk, NA)
      m3.convergence = c(m3.convergence, est$convergence)
    } else {
      est = tmp
      m3.birth.extant = c(m3.birth.extant, est$par[1])
      m3.death.extant = c(m3.death.extant, est$par[2])
      m3.birth.ranges = c(m3.birth.ranges, est$par[3])
      m3.death.ranges = c(m3.death.ranges, est$par[4])
      m3.lk = c(m3.lk, est$value)
      m3.convergence = c(m3.convergence, est$convergence)
    }
  }
  
  cat("finished M3...\n", file=file2, append=T)
  
  ## log output
  
  file1 = paste0("lk_values_", file)
  
  # print headers
  cat(# tree info
    "replicate", "\t",
    "ranges", "\t", # number of ranges
    "max_range", "\t", # max range age. Should = the origin time
    "tips", "\t", # this should be the same for all trees
    "origin", "\t",
    "crown", "\t",
    "rho", "\t",
    # model 1
    "m1_lambda", "\t",
    "m1_mu", "\t",
    "m1_lambda*", "\t",
    "m1_mu*", "\t",
    "m1_lk", "\t",
    "m1_convergence", "\t",
    # model 2
    "m2_lambda", "\t",
    "m2_mu", "\t",
    "m2_lambda*", "\t",
    "m2_mu*", "\t",
    "m2_lk", "\t",
    "m2_convergence", "\t",
    # model 3
    "m3_lambda", "\t",
    "m3_mu", "\t",
    "m3_lambda*", "\t",
    "m3_mu*", "\t",
    "m3_lk", "\t",
    "m3_convergence", "\n",
    sep = "",
    file = file1,
    append = F
  )
  
  for (i in 1:replicates){
    
    tree1<-trees[[i]]
    frs<-ranges[[i]]
    
    # collect useful info
    origin = max(FossilSim::symmetric.ages(tree1, root.edge = T)$start)
    t = geiger::drop.extinct(tree1)
    crown = max(FossilSim::symmetric.ages(t, root.edge = F)$start)
    tips = length(t$tip.label)
    
    cat(# tree info
      i, "\t",
      length(frs$sp), "\t", # number of ranges
      max(frs$start), "\t", # max range age. Should = the origin time
      tips, "\t", # this should be the same for all trees
      origin, "\t",
      crown, "\t",
      rho, "\t",
      # model 1
      m1.birth.extant[i], "\t",
      m1.death.extant[i], "\t",
      m1.birth.ranges[i], "\t",
      m1.death.ranges[i], "\t",
      m1.lk[i], "\t",
      m1.convergence[i], "\t",
      # model 2
      m2.birth.extant[i], "\t",
      m2.death.extant[i], "\t",
      m2.birth.ranges[i], "\t",
      m2.death.ranges[i], "\t",
      m2.lk[i], "\t",
      m2.convergence[i], "\t",
      # model 3
      m3.birth.extant[i], "\t",
      m3.death.extant[i], "\t",
      m3.birth.ranges[i], "\t",
      m3.death.ranges[i], "\t",
      m3.lk[i], "\t",
      m3.convergence[i], "\n",
      sep = "",
      file = file1,
      append = T
    )
    
  }
  
  ## likelihood ratio tests
  
  LR1 = c()
  LR2 = c()
  LR3 = c()
  
  accept1 = c()
  accept2 = c()
  accept3 = c()
  accept4 = c()
  accept5 = c()
  accept6 = c()
  
  if(debugging)
    cat("Model testing...\n")
  
  for(i in 1:replicates){
    
    if(debugging)
      cat(i,"\n")
    
    lk1 = m1.lk[i]
    lk2 = m2.lk[i]
    lk3 = m3.lk[i]
    
    # model 1 (l = l*, m = m*) vs model 2 (l, l*, d), df = 1
    if(is.na(lk1) || is.na(lk2)){
      
      LR1 = c(LR1, NA)
      accept1 = c(accept1, NA)
      accept2 = c(accept2, NA)
      
    } else {
      
      LR = 2 * (lk2 - lk1)
      
      if(LR > ERM_CRM_LR1)
        accept1 = c(accept1, 0) else
          accept1 = c(accept1, 1)
        
      if(LR > ERM_CRM_LR2)
        accept2 = c(accept2, 0) else
          accept2 = c(accept2, 1)  
      
      LR1 = c(LR1, LR)
    }
    
    # model 1 (l = l*, m = m*) vs model 3 (l, l*, m, m*), df = 2
    if(is.na(lk1) || is.na(lk3)){
      
      LR2 = c(LR2, NA)
      accept3 = c(accept3, NA)
      accept4 = c(accept4, NA)
      
    } else {
    
      LR = 2 * (lk3 - lk1)
      
      if(LR > ERM_IRM_LR1)
        accept3 = c(accept3, 0) else
          accept3 = c(accept3, 1)

      if(LR > ERM_IRM_LR2)
        accept4 = c(accept4, 0) else
          accept4 = c(accept4, 1)        
            
      LR2 = c(LR2, LR)
    }
    
    # model 2 (l, l*, d) vs model 3 (l, l*, m, m*), df = 1
    if(is.na(lk2) || is.na(lk3)){
      
      LR3 = c(LR3, NA)
      accept5 = c(accept5, NA)
      accept6 = c(accept6, NA)
      
    } else {
    
      LR = 2 * (lk3 - lk2)
    
      # accept or reject the null model
      # reject the null if:
      if(LR > CRM_IRM_LR1)
        accept5 = c(accept5, 0) else
          accept5 = c(accept5, 1)
        
      if(LR > CRM_IRM_LR2)
        accept6 = c(accept6, 0) else
          accept6 = c(accept6, 1)    
      
      LR3 = c(LR3, LR)
    }
    #eol
  }
  
  cat("finished LR tests...\n", file=file2, append=T)
  
  ## log output - inc. lk ratio test results
  
  file1 = paste0("lk_tests_", file)
  
  if(debugging)
    cat("Printing output...\n")
  
  # print headers
  cat(# tree info
    "replicate", "\t",
    "ranges", "\t", # number of ranges
    "max_range", "\t", # max range age. Should = the origin time
    "tips", "\t", # this should be the same for all trees
    "origin", "\t",
    "crown", "\t",
    "rho", "\t",
    # model 1
    "m1_lambda", "\t",
    "m1_mu", "\t",
    "m1_lambda*", "\t",
    "m1_mu*", "\t",
    "m1_lk", "\t",
    "m1_convergence", "\t",
    # model 2
    "m2_lambda", "\t",
    "m2_mu", "\t",
    "m2_lambda*", "\t",
    "m2_mu*", "\t",
    "m2_lk", "\t",
    "m2_convergence", "\t",
    # model 3
    "m3_lambda", "\t",
    "m3_mu", "\t",
    "m3_lambda*", "\t",
    "m3_mu*", "\t",
    "m3_lk", "\t",
    "m3_convergence", "\t",
    # lk tests
    "LR_1v2", "\t",
    "accept_1v2_95", "\t",
    "accept_1v2_99", "\t",
    "LR_1v3", "\t",
    "accept_1v3_95", "\t",
    "accept_1v3_99", "\t",
    "LR_2v3", "\t",
    "accept_2v3_95", "\t",
    "accept_2v3_99", "\n",
    sep = "",
    file = file1,
    append = F
  )
  
  for (i in 1:replicates){

    tree1<-trees[[i]]
    frs<-ranges[[i]]
        
    # collect useful info
    origin = max(FossilSim::symmetric.ages(tree1, root.edge = T)$start)
    t = geiger::drop.extinct(tree1)
    crown = max(FossilSim::symmetric.ages(t, root.edge = F)$start)
    tips = length(t$tip.label)
    
    cat(# tree info
        i, "\t",
        length(frs$sp), "\t", # number of ranges
        max(frs$start), "\t", # max range age. Should = the origin time
        tips, "\t", # this should be the same for all trees
        origin, "\t",
        crown, "\t",
        rho, "\t",
        # model 1
        m1.birth.extant[i], "\t",
        m1.death.extant[i], "\t",
        m1.birth.ranges[i], "\t",
        m1.death.ranges[i], "\t",
        m1.lk[i], "\t",
        m1.convergence[i], "\t",
        # model 2
        m2.birth.extant[i], "\t",
        m2.death.extant[i], "\t",
        m2.birth.ranges[i], "\t",
        m2.death.ranges[i], "\t",
        m2.lk[i], "\t",
        m2.convergence[i], "\t",
        # model 3
        m3.birth.extant[i], "\t",
        m3.death.extant[i], "\t",
        m3.birth.ranges[i], "\t",
        m3.death.ranges[i], "\t",
        m3.lk[i], "\t",
        m3.convergence[i], "\t",
        # lk tests
        LR1[i], "\t",
        accept1[i], "\t", 
        accept2[i], "\t",
        LR2[i], "\t",
        accept3[i], "\t",
        accept4[i], "\t",
        LR3[i], "\t",
        accept5[i], "\t",
        accept6[i], "\n",
        sep = "",
        file = file1,
        append = T
    )
    
  }
  
  ## expected values
  
  if(shifts){
    lb = c()
    mb = c()
    ls = c()
    ms = c()
    
    for(i in 1:replicates){
      tree1 = trees[[i]]
      origin = max(FossilSim::symmetric.ages(tree1, root.edge = T)$start)
      
      # DOUBLE CHECK THESE CALCULATIONS ARE CORRECT
      if(shifts){
        if(origin < 2){
          v1 = lambda.branching # value 1 etc
          v2 = mu.branching
          v3 = lambda.star
          v4 = mu.star
        } else if(origin < 4){
          v1 = ((origin - 2)/origin) * 1.5 + (2/origin) * lambda.branching
          v2 = ((origin - 2)/origin) * 0 + (2/origin) * mu.branching
          v3 = (1-f) * 2 * v1 + f * v1 + lambda.a
          v4 = (1-f) * v1 + lambda.a + v2
        } else {
          v1 = ((origin - 2)/origin) * 1.5 + (2/origin) * lambda.branching + (origin - 4) * lambda.branching
          v2 = ((origin - 2)/origin) * 0 + (2/origin) * mu.branching + (origin - 4) * mu.branching
          v3 = (1-f) * 2 * v1 + f * v1 + lambda.a
          v4 = (1-f) * v1 + lambda.a + v2
        }
        
        lb = c(lb, v1)
        mb = c(mb, v2)
        
        ls = c(ls, v3)
        ms = c(ms, v4)
      }
    }
    
  } else {
    lb = lambda.branching
    mb = mu.branching
    # NEW PARAMETERISATION
    ls = f*2*lambda.branching+(1-f)*lambda.branching+lambda.a
    ms = f*lambda.branching+lambda.a+mu.branching
  }
    
  return(list(lb = lb, mb = mb, ls = ls, ms = ms,
              # model 1
              m1.birth.extant = m1.birth.extant,
              m1.death.extant = m1.death.extant,
              m1.birth.ranges = m1.birth.ranges,
              m1.death.ranges = m1.death.ranges,
              m1.lk = m1.lk,
              m1.convergence = m1.convergence,
              # model 2
              m2.birth.extant = m2.birth.extant,
              m2.death.extant = m2.death.extant,
              m2.birth.ranges = m2.birth.ranges,
              m2.death.ranges = m2.death.ranges,
              m2.lk = m2.lk,
              m2.convergence = m2.convergence,
              # model 3
              m3.birth.extant = m3.birth.extant,
              m3.death.extant = m3.death.extant,
              m3.birth.ranges = m3.birth.ranges,
              m3.death.ranges = m3.death.ranges,
              m3.lk = m3.lk,
              m3.convergence = m3.convergence,
              LR1 = LR1, LR2 = LR2, LR3 = LR3,
              accept1 = accept1, accept2 = accept2, accept3 = accept3,
              accept4 = accept4, accept5 = accept5, accept6 = accept6
  ))
  #eof
}

##### # figure 1 - estimates

estimates.figure<-function(input, lb1, lb2, shifts = FALSE, lb3 = NULL, xylim = NULL){

  if( (shifts == TRUE) && (is.null(lb3)) )
    stop("Specify lb3 for shifts == TRUE")

  # parameters
  b = lambda.branching
  d = mu.branching

  # PREVIOUS PARAMETERISATION
  # expected estimate of lambda
  #lambda.star = (1-f)*2*lambda.branching+f*lambda.branching+lambda.a
  # expected estimate of mu
  #mu.star = (1-f)*lambda.branching+lambda.a+mu.branching
  
  # NEW PARAMETERISATION
  if(!exists("lambda.star"))
    lambda.star = f*2*lambda.branching+(1-f)*lambda.branching+lambda.a
  if(!exists("mu.star"))
    mu.star = f*lambda.branching+lambda.a+mu.branching

  birth.ranges = as.vector(na.omit(input$m3.birth.ranges))
  death.ranges = as.vector(na.omit(input$m3.death.ranges))

  birth.extant = as.vector(na.omit(input$m3.birth.extant))
  death.extant = as.vector(na.omit(input$m3.death.extant))
  
  if(is.null(xylim))
    xylim = 0.7

  if(shifts){
    p = ggplot() +
      geom_point(aes(birth.extant,death.extant),colour = "tomato") +
      geom_point(aes(birth.ranges,death.ranges),colour = "dodgerblue3") +
      #geom_point(aes(b,d), shape=17, size=4) +
      #geom_point(aes(lambda.star,mu.star), shape=8, size=4) +
      ylim(0,xylim) + xlim(0,xylim) +
      xlab("Speciation rate") + ylab("Extinction rate") +
      annotate("text",0,xylim,label=lb1,size=4,parse=T,hjust = 0) +
      annotate("text",0,xylim*0.9,label=lb3,size=4,parse=T,hjust = 0) +
      annotate("text",0,xylim*0.8,label=lb2,size=4,parse=T,hjust = 0)
  } else {
    p = ggplot() +
      geom_point(aes(birth.extant,death.extant),colour = "tomato") +
      geom_point(aes(birth.ranges,death.ranges),colour = "dodgerblue3") +
      geom_point(aes(b,d), shape=17, size=4) +
      geom_point(aes(lambda.star,mu.star), shape=8, size=4) +
      ylim(0,xylim) + xlim(0,xylim) +
      xlab("Speciation rate") + ylab("Extinction rate") +
      annotate("text",0,xylim,label=lb1,size=4,parse=T,hjust = 0) +
      annotate("text",0,xylim*0.9,label=lb2,size=4,parse=T,hjust = 0)
      if(!(is.null(lb3)))
        p = p + annotate("text",0,xylim*0.8,label=lb3,size=4,parse=T,hjust = 0)
  }
  return(p)
}

##### # figure 2 - correlations

correlations.figure<-function(input, lb1, lb2, equation = 1, shifts = FALSE, lb3 = NULL){

  if( (shifts == TRUE) && (is.null(lb3)) )
    stop("Specify lb3 for shifts == TRUE")

  # parameters
  b = lambda.branching
  d = mu.branching

  # expected estimate of lambda
  #lambda.star = (1-f)*2*lambda.branching+f*lambda.branching+lambda.a
  # expected estimate of mu
  #mu.star = (1-f)*lambda.branching+lambda.a+mu.branching

  if(equation == 1){
    xa = input$m3.birth.extant - input$m3.death.extant
    ya = input$m3.birth.ranges - input$m3.death.ranges
  } else if (equation == 2) {
    xa = input$m3.birth.ranges - (input$m3.birth.extant * 2)
    ya = input$m3.death.ranges - (input$m3.birth.extant + input$m3.death.extant)
  } else if (equation == 3) {
    xa = input$m3.birth.ranges - input$m3.birth.extant
    ya = input$m3.death.ranges - input$m3.death.extant
  }
  
  xa = as.vector(na.omit(xa))
  ya = as.vector(na.omit(ya))

  df = data.frame(xa=xa,ya=ya)

  r2 = round(summary(lm(xa~ya))$r.squared,2)

  lb.R = paste("R^2 == ", r2, sep="")

  y2 = max(ya) - (abs(max(ya)-min(ya))*0.1)

  # shifts + cryptic species only
  y3 = max(ya) - (abs(max(ya)-min(ya))*0.2)

  if(shifts){
    p = ggplot(df,aes(xa,ya)) +
      geom_point(colour="cadetblue3") +
      geom_smooth(method=lm,se=F,colour="black",size=0.5) +
      annotate("text",min(xa),max(ya),label=lb1,size=4,parse=T,hjust = 0) +
      annotate("text",min(xa),y2,label=lb3,size=4,parse=T,hjust = 0) +
      annotate("text",min(xa),y3,label=lb2,size=4,parse=T,hjust = 0) +
      annotate("text",max(xa),min(ya),label=lb.R,size=4,parse=T,hjust = 1)
  } else {
    p = ggplot(df,aes(xa,ya)) +
      geom_point(colour="cadetblue3") +
      geom_smooth(method=lm,se=F,colour="black",size=0.5) +
      annotate("text",min(xa),max(ya),label=lb1,size=4,parse=T,hjust = 0) +
      annotate("text",min(xa),y2,label=lb2,size=4,parse=T,hjust = 0) +
      annotate("text",max(xa),min(ya),label=lb.R,size=4,parse=T,hjust = 1)
    if(!(is.null(lb3)))
      p = p + annotate("text",min(xa),y3,label=lb3,size=4,parse=T,hjust = 0)
  }

  if(equation == 1)
    p = p + xlab(expression(paste(lambda,"-",mu))) + ylab(expression(paste(lambda,"*-",mu,"*")))
  else if(equation == 2)
    p = p + xlab(expression(paste(lambda,"*-",2,lambda))) + ylab(expression(paste(mu,"*-(",lambda,"+",mu,")")))
  else if(equation == 3)
    p = p + xlab(expression(paste(lambda,"*-",lambda))) + ylab(expression(paste(mu,"*-",mu)))
    
  return(p)

}

#### # check for convergence issues

convergence.issues<-function(input, file = ""){

  # the way the code is set up, the following scenario should be avoided
  if(any(input$m1.convergence != 0)){
    rep = which(input$m1.convergence != 0)
    cat("WARNING: there appear to be some convergence issues with model 1, replicates", rep, "\n", file = file, sep = " ", append = T)
  }
  if(any(input$m2.convergence != 0)){
    rep = which(input$m2.convergence != 0)
    cat("WARNING: there appear to be some convergence issues with model 2, replicates", rep, "\n", file = file, sep = " ", append = T)
  }
  if(any(input$m3.convergence != 0)){
    rep = which(input$m3.convergence != 0)
    cat("WARNING: there appear to be some convergence issues with model 3, replicates", rep, "\n", file = file, sep = " ", append = T)
  }
}

##### # table 1 - estimates & likelihood ratio test

print.headers<-function(file = "lk_ratio_test_results_simulations.txt"){

  cat("lambda", "\t",
      "mu", "\t",
      "lambda*", "\t",
      "mu*", "\t",
      "d", "\t",
      "d*", "\t",
      # model 1
      "m1_l","\t",
      "m1_m","\t",
      "m1_l*","\t",
      "m1_m*","\t",
      "m1_d","\t",
      "m1_d*","\t",
      # model 2
      "m2_l","\t",
      "m2_m","\t",
      "m2_l*","\t",
      "m2_m*","\t",
      "m2_d","\t",
      "m2_d*","\t",
      # model 3
      "m3_l","\t",
      "m3_m","\t",
      "m3_l*","\t",
      "m3_m*","\t",
      "m3_d","\t",
      "m3_d*","\t",
      # lk ratio
      "m1_v_m2_95","\t",
      "m1_v_m2_99","\t",
      "m1_v_m3_95","\t",
      "m1_v_m3_99","\t",
      "m2_v_m3_95","\t",
      "m2_v_m3_99","\n", # this is the original test
      file = file, sep = "",
      append = F
  )
#eof
}

print.results<-function(p, file = "lk_ratio_test_results_simulations.txt"){

  cat(round(mean(p$lb, na.rm = T),3), "\t",
      round(mean(p$mb, na.rm = T),3), "\t",
      round(mean(p$ls, na.rm = T),3), "\t",
      round(mean(p$ms, na.rm = T),3), "\t",
      round(mean(p$lb - p$mb, na.rm = T),3), "\t",
      round(mean(p$ls - p$ms, na.rm = T),3), "\t",
      # model 1
      round(mean(p$m1.birth.extant, na.rm = T),3),"\t",
      round(mean(p$m1.death.extant, na.rm = T),3),"\t",
      round(mean(p$m1.birth.ranges, na.rm = T),3),"\t",
      round(mean(p$m1.death.ranges, na.rm = T),3),"\t",
      round(mean(p$m1.birth.extant - p$m1.death.extant, na.rm = T),3), "\t",
      round(mean(p$m1.birth.ranges - p$m1.death.ranges, na.rm = T),3), "\t",
      # model 2
      round(mean(p$m2.birth.extant, na.rm = T),3),"\t",
      round(mean(p$m2.death.extant, na.rm = T),3),"\t",
      round(mean(p$m2.birth.ranges, na.rm = T),3),"\t",
      round(mean(p$m2.death.ranges, na.rm = T),3),"\t",
      round(mean(p$m2.birth.extant - p$m2.death.extant, na.rm = T),3), "\t",
      round(mean(p$m2.birth.ranges - p$m2.death.ranges, na.rm = T),3), "\t",
      # model 3
      round(mean(p$m3.birth.extant, na.rm = T),3),"\t",
      round(mean(p$m3.death.extant, na.rm = T),3),"\t",
      round(mean(p$m3.birth.ranges, na.rm = T),3),"\t",
      round(mean(p$m3.death.ranges, na.rm = T),3),"\t",
      round(mean(p$m3.birth.extant - p$m3.death.extant, na.rm = T),3), "\t",
      round(mean(p$m3.birth.ranges - p$m3.death.ranges, na.rm = T),3), "\t",
      # lk ratio
      round(sum(p$accept1, na.rm = T)/length(na.omit(p$accept1)),3),"\t",
      round(sum(p$accept2, na.rm = T)/length(na.omit(p$accept2)),3),"\t",
      round(sum(p$accept3, na.rm = T)/length(na.omit(p$accept3)),3),"\t",
      round(sum(p$accept4, na.rm = T)/length(na.omit(p$accept4)),3),"\t",
      round(sum(p$accept5, na.rm = T)/length(na.omit(p$accept5)),3),"\t",
      round(sum(p$accept6, na.rm = T)/length(na.omit(p$accept6)),3),"\n",
      file = file, sep = "",
      append = T
  )

 # eof
}

##### print data

print.trees<-function(trees,prefix){
  for (i in 1:length(trees)){
    t <- trees[[i]]
    ape::write.tree(t, file = paste("data/t",prefix,".",i,".txt",sep=""))
  }
  #eof
}

print.ranges<-function(ranges,prefix){
  for (i in 1:length(ranges)){
    r <- ranges[[i]]
    write.table(r, file = paste("data/o",prefix,".",i,".txt",sep=""))
  }
  #eof
}
